import isFunction from "../../function/isFunction";
var Identity = (function () {
    function Identity(value) {
        if (value === undefined)
            throw new Error("Identity: Value is undefined");
        this._value = value;
    }
    Identity.of = function (v) {
        return new Identity(v);
    };
    Identity.prototype.of = function (v) {
        return new Identity(v);
    };
    Identity.prototype.getValue = function () {
        return this._value;
    };
    Identity.prototype.equals = function (j) {
        return this.getValue() === j.getValue();
    };
    Identity.prototype.concat = function (i) {
        var v = i.getValue();
        if (typeof v !== typeof this.getValue())
            throw new Error("Identity: types do not match to concat");
        if (!v.concat || !this.getValue().concat)
            throw new Error("Identity: Semigroup required to concat");
        return this.map(function (x) { return x.concat(v); });
    };
    Identity.prototype.ap = function (i) {
        var fn = this.getValue();
        if (!isFunction(fn))
            throw Error("Identity: Function not found");
        return new Identity(fn(i.getValue()));
    };
    Identity.prototype.map = function (f) {
        if (!isFunction(f))
            throw Error("Identity: Expected a function");
        return new Identity(f(this.getValue()));
    };
    Identity.prototype.chain = function (f) {
        if (!isFunction(f))
            throw Error("Identity: Expected a function");
        var res = f(this.getValue());
        if (!(res instanceof Identity))
            throw Error("Identity: Function must return Identity");
        return res;
    };
    Identity.prototype.toString = function () {
        var v = this.getValue();
        return "Identity(" + v + ")";
    };
    return Identity;
}());
Identity.prototype["fantasy-land/equals"] = Identity.prototype.equals;
Identity.prototype["fantasy-land/map"] = Identity.prototype.map;
Identity.prototype["fantasy-land/concat"] = Identity.prototype.concat;
Identity.prototype["fantasy-land/chain"] = Identity.prototype.chain;
Identity.prototype["fantasy-land/of"] = Identity.prototype.of;
Identity.prototype["fantasy-land/ap"] = Identity.prototype.ap;
export default Identity;
