import { __extends } from "tslib";
import caseOf from "../../function/caseOf";
import curry from "../../function/curry";
import isFunction from "../../function/isFunction";
var Maybe = (function () {
    function Maybe() {
    }
    Maybe.of = function (v) {
        return new Just(v);
    };
    Maybe.zero = function () {
        return new Nothing();
    };
    Maybe.Just = function (v) {
        return new Just(v);
    };
    Maybe.Nothing = function () {
        return new Nothing();
    };
    Maybe.fromNullable = function (v) {
        return v !== undefined && v !== null ? new Just(v) : new Nothing();
    };
    Maybe.withDefault = function (def, v) {
        return v !== undefined && v !== null ? new Just(v) : new Just(def);
    };
    Maybe.catMaybes = function (ar) {
        return ar.filter(function (m) { return m.isJust(); }).map(function (m) { return m.getValue(); });
    };
    Maybe.isNothing = function (v) {
        return v.isNothing();
    };
    Maybe.isJust = function (v) {
        return v.isJust();
    };
    Maybe.prototype.of = function (v) {
        return new Just(v);
    };
    Maybe.prototype.alt = function (v) {
        var _this = this;
        return caseOf({
            Nothing: function (x) { return v; },
            Just: function (x) { return _this; }
        }, this);
    };
    Maybe.prototype.ap = function (j) {
        return this.caseOf({
            Nothing: function (v) { return j; },
            Just: function (v) {
                if (!isFunction(v))
                    throw Error("Either: Wrapped value is not a function");
                return j.map(v);
            }
        });
    };
    Maybe.prototype.getValue = function () {
        return this._value;
    };
    return Maybe;
}());
export default Maybe;
Maybe.prototype["fantasy-land/equals"] = Maybe.prototype.equals;
Maybe.prototype["fantasy-land/map"] = Maybe.prototype.map;
Maybe.prototype["fantasy-land/chain"] = Maybe.prototype.chain;
Maybe.prototype["fantasy-land/of"] = Maybe.prototype.of;
Maybe.prototype["fantasy-land/zero"] = Maybe.prototype.zero;
Maybe.prototype["fantasy-land/ap"] = Maybe.prototype.ap;
var Nothing = (function (_super) {
    __extends(Nothing, _super);
    function Nothing() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    Nothing.prototype.equals = function (n) {
        return n instanceof Nothing && n.isNothing && n.isNothing();
    };
    Nothing.prototype.map = function (f) {
        return new Nothing();
    };
    Nothing.prototype.chain = function (f) {
        return new Nothing();
    };
    Nothing.prototype.isJust = function () {
        return false;
    };
    Nothing.prototype.isNothing = function () {
        return true;
    };
    Nothing.prototype.toString = function () {
        return "Nothing()";
    };
    Nothing.prototype.caseOf = function (o) {
        if (o.Nothing)
            return o.Nothing();
        else
            throw Error("Maybe: Expected Nothing!");
    };
    return Nothing;
}(Maybe));
export { Nothing };
var Just = (function (_super) {
    __extends(Just, _super);
    function Just(v) {
        var _this = _super.call(this) || this;
        _this._value = v;
        return _this;
    }
    Just.prototype.equals = function (j) {
        return (j instanceof Just &&
            j.isJust &&
            j.isJust() &&
            j.getValue() === this.getValue());
    };
    Just.prototype.map = function (f) {
        if (!isFunction(f))
            throw Error("Maybe: Expected a function");
        return new Just(f(this.getValue()));
    };
    Just.prototype.chain = function (f) {
        if (!isFunction(f))
            throw Error("Maybe: Expected a function");
        return f(this.getValue());
    };
    Just.prototype.isJust = function () {
        return true;
    };
    Just.prototype.isNothing = function () {
        return false;
    };
    Just.prototype.toString = function () {
        return "Just(" + this.getValue() + ")";
    };
    Just.prototype.caseOf = function (o) {
        if (o.Just)
            return o.Just(this.getValue());
        else
            throw Error("Maybe: Expected Just");
    };
    return Just;
}(Maybe));
export { Just };
export var eitherToMaybe = function (e) {
    return caseOf({
        Left: function (v) { return Maybe.Nothing(); },
        Right: function (v) { return Maybe.Just(v); }
    }, e);
};
var _prop = function (key, o) {
    if (typeof key !== "string" && typeof key !== "number")
        throw Error("Key should be either string or number");
    if (!o)
        return Maybe.Nothing();
    return key in o ? Maybe.fromNullable(o[key]) : Maybe.Nothing();
};
export var prop = curry(_prop);
