import isFunction from "../../function/isFunction";
var Pair = (function () {
    function Pair(v1, v2) {
        if (v1 === undefined || v2 === undefined)
            throw Error("Pair: Both first and second values must be defined");
        this._value = [v1, v2];
    }
    Pair.of = function (v) {
        return new Pair(v, v);
    };
    Pair.prototype.of = function (v) {
        return new Pair(v, v);
    };
    Pair.prototype.equals = function (j) {
        return j.fst() === this.fst() && j.snd() === this.snd();
    };
    Pair.prototype.concat = function (p) {
        if (!(p instanceof Pair))
            throw Error("Pair: Pair required");
        var lf = this.fst();
        var ls = this.snd();
        var rf = p.fst();
        var rs = p.snd();
        if (!lf.concat || !ls.concat || !rf.concat || !rs.concat)
            throw Error("Pair: Both Pairs must contain Semigroups");
        return new Pair(lf.concat(rf), ls.concat(rs));
    };
    Pair.prototype.fst = function () {
        return this.getValue()[0];
    };
    Pair.prototype.snd = function () {
        return this.getValue()[1];
    };
    Pair.prototype.ap = function (j) {
        if (!(j instanceof Pair))
            throw Error("Pair: Pair required");
        var fn = this.snd();
        if (!isFunction(fn))
            throw Error("Pair: Second wrapped value should be a function");
        var l = this.fst();
        var r = j.fst();
        if (!l.concat || !r.concat)
            throw Error("Pair: Types should be Semigroups");
        return new Pair(l.concat(r), fn(r));
    };
    Pair.prototype.getValue = function () {
        return this._value;
    };
    Pair.prototype.map = function (f) {
        if (!isFunction(f))
            throw Error("Pair: Expected a function");
        return new Pair(this.fst(), f(this.snd()));
    };
    Pair.prototype.bimap = function (f1, f2) {
        if (!isFunction(f1) || !isFunction(f2))
            throw Error("Pair: Expected functions for both parts");
        return new Pair(f1(this.fst()), f2(this.snd()));
    };
    Pair.prototype.chain = function (f) {
        if (!isFunction(f))
            throw Error("Pair: Expected a function");
        var l = this.fst();
        if (!l.concat)
            throw Error("Pair: First value should be a Semigroup");
        var p = f(this.snd());
        if (!(p instanceof Pair))
            throw Error("Pair: Function must return a Pair");
        var r = p.fst();
        if (!r.concat)
            throw Error("Pair: First value of the returned Pair should be a Semigroup");
        return new Pair(l.concat(r), p.snd());
    };
    Pair.prototype.swap = function () {
        var v = this.getValue();
        return new Pair(v[1], v[0]);
    };
    Pair.prototype.toString = function () {
        var v = this.getValue();
        return "Pair((" + v[0] + "), (" + v[1] + "))";
    };
    return Pair;
}());
Pair.prototype["fantasy-land/equals"] = Pair.prototype.equals;
Pair.prototype["fantasy-land/map"] = Pair.prototype.map;
Pair.prototype["fantasy-land/concat"] = Pair.prototype.concat;
Pair.prototype["fantasy-land/bimap"] = Pair.prototype.bimap;
Pair.prototype["fantasy-land/chain"] = Pair.prototype.chain;
Pair.prototype["fantasy-land/of"] = Pair.prototype.of;
Pair.prototype["fantasy-land/ap"] = Pair.prototype.ap;
export default Pair;
