import isFunction from "../../function/isFunction";
import Pair from "../Pair";
var State = (function () {
    function State(fn) {
        if (!isFunction(fn))
            throw Error("State: Must wrap a function");
        this._value = fn;
    }
    State.of = function (v) {
        return new State(function (s) { return Pair(v, s); });
    };
    State.prototype.of = function (v) {
        return new State(function (s) { return Pair(v, s); });
    };
    State.prototype.get = function (fn) {
        if (!fn)
            return new State(function (s) { return Pair(s, s); });
        if (isFunction(fn))
            return new State(function (s) { return Pair(fn(s), s); });
        throw Error("State: No argument or function required");
    };
    State.prototype.put = function (s) {
        if (s)
            return new State(function (_) { return Pair(s, _); });
        throw Error("State: No argument or function required");
    };
    State.prototype.runWith = function (s) {
        var p = this._value(s);
        if (!p.fst || !p.snd)
            throw Error("State: Must wrap a function of the form s -> Pair a s");
        return p;
    };
    State.prototype.execWith = function (s) {
        var pair = this.runWith(s);
        return pair.snd();
    };
    State.prototype.evalWith = function (s) {
        var pair = this.runWith(s);
        return pair.fst();
    };
    State.prototype.ap = function (j) {
        if (!(j instanceof State))
            throw Error("State: State required");
        var runWith = this._value;
        return new State(function (e) {
            var p = runWith(e);
            if (!p.fst || !p.snd)
                throw Error("State: Must wrap a function of the form s -> Pair a s");
            var fn = p.fst();
            if (!isFunction(fn))
                throw Error("State: Source value must be a function");
            return j.map(fn).runWith(p.snd());
        });
    };
    State.prototype.getValue = function () {
        return this._value;
    };
    State.prototype.map = function (f) {
        if (!isFunction(f))
            throw Error("State: Expected a function");
        var runWith = this._value;
        return new State(function (s) {
            var m = runWith(s);
            return Pair(f(m.fst()), m.snd());
        });
    };
    State.prototype.chain = function (f) {
        if (!isFunction(f))
            throw Error("State: Expected a function");
        var runWith = this.getValue();
        return new State(function (s) {
            var p = runWith(s);
            var m = f(p.fst());
            if (!(m instanceof State))
                throw Error("State: Function must return a State");
            return m.runWith(p.snd());
        });
    };
    State.prototype.toString = function () {
        var v = this.getValue();
        return "State(" + v + ")";
    };
    return State;
}());
State.prototype["fantasy-land/map"] = State.prototype.map;
State.prototype["fantasy-land/chain"] = State.prototype.chain;
State.prototype["fantasy-land/of"] = State.prototype.of;
State.prototype["fantasy-land/ap"] = State.prototype.ap;
export default State;
