'use strict';

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

function __spreadArrays() {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
}

function isFunction(f) {
    return typeof f === "function";
}

function curry(fn) {
    if (!isFunction(fn)) throw Error("Function not provided");
    var slice = Array.prototype.slice;
    var arity = fn.length;
    return function curried() {
        var args = slice.call(arguments, 0);
        if (args.length >= arity) return fn.apply(null, args);
        return function () {
            return curried.apply(null, args.concat(slice.call(arguments)));
        };
    };
}

function caseOf(o, p) {
    if (isFunction(p.caseOf)) return p.caseOf(o);else throw Error("Unable to pattern match");
}
var caseOf$1 = curry(caseOf);

var Either = function () {
    function Either() {}
    Either.of = function (v) {
        return new Right(v);
    };
    Either.Right = function (v) {
        return new Right(v);
    };
    Either.Left = function (v) {
        return new Left(v);
    };
    Either.fromNullable = function (v) {
        return v !== null || v !== undefined ? new Right(v) : new Left(v);
    };
    Either.withDefault = function (def, v) {
        return v.caseOf({ Left: function Left() {
                return Either.of(def);
            }, Right: function Right() {
                return v;
            } });
    };
    Either.swap = function (e) {
        return e.swap();
    };
    Either.try = function (f) {
        return function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            try {
                return new Right(f.apply(null, args));
            } catch (e) {
                return new Left(e);
            }
        };
    };
    Either.bimap = function (e, fl, fr) {
        return e.bimap(fl, fr);
    };
    Either.isLeft = function (v) {
        return v.isLeft();
    };
    Either.isRight = function (v) {
        return v.isRight();
    };
    Either.prototype.of = function (v) {
        return new Right(v);
    };
    Either.prototype.ap = function (j) {
        return this.caseOf({
            Left: function Left(v) {
                return j;
            },
            Right: function Right(v) {
                if (!isFunction(v)) throw Error("Either: Wrapped value is not a function");
                return j.map(v);
            }
        });
    };
    Either.prototype.getValue = function () {
        return this._value;
    };
    return Either;
}();
Either.prototype["fantasy-land/equals"] = Either.prototype.equals;
Either.prototype["fantasy-land/map"] = Either.prototype.map;
Either.prototype["fantasy-land/bimap"] = Either.prototype.bimap;
Either.prototype["fantasy-land/chain"] = Either.prototype.chain;
Either.prototype["fantasy-land/of"] = Either.prototype.of;
Either.prototype["fantasy-land/ap"] = Either.prototype.ap;
var Left = function (_super) {
    __extends(Left, _super);
    function Left(v) {
        var _this = _super.call(this) || this;
        _this._value = v;
        return _this;
    }
    Left.prototype.equals = function (n) {
        return n instanceof Left && n.isLeft && n.isLeft() && n.getValue() === this.getValue();
    };
    Left.prototype.map = function (f) {
        return new Left(this.getValue());
    };
    Left.prototype.bimap = function (fl, fr) {
        return new Left(fl(this.getValue()));
    };
    Left.prototype.chain = function (f) {
        return new Left(this.getValue());
    };
    Left.prototype.isRight = function () {
        return false;
    };
    Left.prototype.isLeft = function () {
        return true;
    };
    Left.prototype.swap = function () {
        return new Right(this.getValue());
    };
    Left.prototype.toString = function () {
        return "Left(" + this.getValue() + ")";
    };
    Left.prototype.caseOf = function (o) {
        if (o.Left) return o.Left(this.getValue());else throw Error("Either: Expected Left!");
    };
    return Left;
}(Either);
var Right = function (_super) {
    __extends(Right, _super);
    function Right(v) {
        var _this = _super.call(this) || this;
        _this._value = v;
        return _this;
    }
    Right.prototype.equals = function (j) {
        return j instanceof Right && j.isRight && j.isRight() && j.getValue() === this.getValue();
    };
    Right.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("Either: Expected a function to map");
        return new Right(f(this.getValue()));
    };
    Right.prototype.bimap = function (fl, fr) {
        return new Right(fr(this.getValue()));
    };
    Right.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("Either: Expected a function to chain");
        return f(this.getValue());
    };
    Right.prototype.isRight = function () {
        return true;
    };
    Right.prototype.isLeft = function () {
        return false;
    };
    Right.prototype.swap = function () {
        return new Left(this.getValue());
    };
    Right.prototype.toString = function () {
        return "Right(" + this.getValue() + ")";
    };
    Right.prototype.caseOf = function (o) {
        if (o.Right) return o.Right(this.getValue());else throw Error("Either: Expected Right");
    };
    return Right;
}(Either);
var maybeToEither = function maybeToEither(m) {
    return caseOf$1({
        Nothing: function Nothing(v) {
            return Either.Left(null);
        },
        Just: function Just(v) {
            return Either.Right(v);
        }
    }, m);
};

var Maybe = function () {
    function Maybe() {}
    Maybe.of = function (v) {
        return new Just(v);
    };
    Maybe.zero = function () {
        return new Nothing();
    };
    Maybe.Just = function (v) {
        return new Just(v);
    };
    Maybe.Nothing = function () {
        return new Nothing();
    };
    Maybe.fromNullable = function (v) {
        return v !== undefined && v !== null ? new Just(v) : new Nothing();
    };
    Maybe.withDefault = function (def, v) {
        return v !== undefined && v !== null ? new Just(v) : new Just(def);
    };
    Maybe.catMaybes = function (ar) {
        return ar.filter(function (m) {
            return m.isJust();
        }).map(function (m) {
            return m.getValue();
        });
    };
    Maybe.isNothing = function (v) {
        return v.isNothing();
    };
    Maybe.isJust = function (v) {
        return v.isJust();
    };
    Maybe.prototype.of = function (v) {
        return new Just(v);
    };
    Maybe.prototype.alt = function (v) {
        var _this = this;
        return caseOf$1({
            Nothing: function Nothing(x) {
                return v;
            },
            Just: function Just(x) {
                return _this;
            }
        }, this);
    };
    Maybe.prototype.ap = function (j) {
        return this.caseOf({
            Nothing: function Nothing(v) {
                return j;
            },
            Just: function Just(v) {
                if (!isFunction(v)) throw Error("Either: Wrapped value is not a function");
                return j.map(v);
            }
        });
    };
    Maybe.prototype.getValue = function () {
        return this._value;
    };
    return Maybe;
}();
Maybe.prototype["fantasy-land/equals"] = Maybe.prototype.equals;
Maybe.prototype["fantasy-land/map"] = Maybe.prototype.map;
Maybe.prototype["fantasy-land/chain"] = Maybe.prototype.chain;
Maybe.prototype["fantasy-land/of"] = Maybe.prototype.of;
Maybe.prototype["fantasy-land/zero"] = Maybe.prototype.zero;
Maybe.prototype["fantasy-land/ap"] = Maybe.prototype.ap;
var Nothing = function (_super) {
    __extends(Nothing, _super);
    function Nothing() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    Nothing.prototype.equals = function (n) {
        return n instanceof Nothing && n.isNothing && n.isNothing();
    };
    Nothing.prototype.map = function (f) {
        return new Nothing();
    };
    Nothing.prototype.chain = function (f) {
        return new Nothing();
    };
    Nothing.prototype.isJust = function () {
        return false;
    };
    Nothing.prototype.isNothing = function () {
        return true;
    };
    Nothing.prototype.toString = function () {
        return "Nothing()";
    };
    Nothing.prototype.caseOf = function (o) {
        if (o.Nothing) return o.Nothing();else throw Error("Maybe: Expected Nothing!");
    };
    return Nothing;
}(Maybe);
var Just = function (_super) {
    __extends(Just, _super);
    function Just(v) {
        var _this = _super.call(this) || this;
        _this._value = v;
        return _this;
    }
    Just.prototype.equals = function (j) {
        return j instanceof Just && j.isJust && j.isJust() && j.getValue() === this.getValue();
    };
    Just.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("Maybe: Expected a function");
        return new Just(f(this.getValue()));
    };
    Just.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("Maybe: Expected a function");
        return f(this.getValue());
    };
    Just.prototype.isJust = function () {
        return true;
    };
    Just.prototype.isNothing = function () {
        return false;
    };
    Just.prototype.toString = function () {
        return "Just(" + this.getValue() + ")";
    };
    Just.prototype.caseOf = function (o) {
        if (o.Just) return o.Just(this.getValue());else throw Error("Maybe: Expected Just");
    };
    return Just;
}(Maybe);
var eitherToMaybe = function eitherToMaybe(e) {
    return caseOf$1({
        Left: function Left(v) {
            return Maybe.Nothing();
        },
        Right: function Right(v) {
            return Maybe.Just(v);
        }
    }, e);
};
var _prop = function _prop(key, o) {
    if (typeof key !== "string" && typeof key !== "number") throw Error("Key should be either string or number");
    if (!o) return Maybe.Nothing();
    return key in o ? Maybe.fromNullable(o[key]) : Maybe.Nothing();
};
var prop = curry(_prop);

var Pair = function () {
    function Pair(v1, v2) {
        if (v1 === undefined || v2 === undefined) throw Error("Pair: Both first and second values must be defined");
        this._value = [v1, v2];
    }
    Pair.of = function (v) {
        return new Pair(v, v);
    };
    Pair.prototype.of = function (v) {
        return new Pair(v, v);
    };
    Pair.prototype.equals = function (j) {
        return j.fst() === this.fst() && j.snd() === this.snd();
    };
    Pair.prototype.concat = function (p) {
        if (!(p instanceof Pair)) throw Error("Pair: Pair required");
        var lf = this.fst();
        var ls = this.snd();
        var rf = p.fst();
        var rs = p.snd();
        if (!lf.concat || !ls.concat || !rf.concat || !rs.concat) throw Error("Pair: Both Pairs must contain Semigroups");
        return new Pair(lf.concat(rf), ls.concat(rs));
    };
    Pair.prototype.fst = function () {
        return this.getValue()[0];
    };
    Pair.prototype.snd = function () {
        return this.getValue()[1];
    };
    Pair.prototype.ap = function (j) {
        if (!(j instanceof Pair)) throw Error("Pair: Pair required");
        var fn = this.snd();
        if (!isFunction(fn)) throw Error("Pair: Second wrapped value should be a function");
        var l = this.fst();
        var r = j.fst();
        if (!l.concat || !r.concat) throw Error("Pair: Types should be Semigroups");
        return new Pair(l.concat(r), fn(r));
    };
    Pair.prototype.getValue = function () {
        return this._value;
    };
    Pair.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("Pair: Expected a function");
        return new Pair(this.fst(), f(this.snd()));
    };
    Pair.prototype.bimap = function (f1, f2) {
        if (!isFunction(f1) || !isFunction(f2)) throw Error("Pair: Expected functions for both parts");
        return new Pair(f1(this.fst()), f2(this.snd()));
    };
    Pair.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("Pair: Expected a function");
        var l = this.fst();
        if (!l.concat) throw Error("Pair: First value should be a Semigroup");
        var p = f(this.snd());
        if (!(p instanceof Pair)) throw Error("Pair: Function must return a Pair");
        var r = p.fst();
        if (!r.concat) throw Error("Pair: First value of the returned Pair should be a Semigroup");
        return new Pair(l.concat(r), p.snd());
    };
    Pair.prototype.swap = function () {
        var v = this.getValue();
        return new Pair(v[1], v[0]);
    };
    Pair.prototype.toString = function () {
        var v = this.getValue();
        return "Pair((" + v[0] + "), (" + v[1] + "))";
    };
    return Pair;
}();
Pair.prototype["fantasy-land/equals"] = Pair.prototype.equals;
Pair.prototype["fantasy-land/map"] = Pair.prototype.map;
Pair.prototype["fantasy-land/concat"] = Pair.prototype.concat;
Pair.prototype["fantasy-land/bimap"] = Pair.prototype.bimap;
Pair.prototype["fantasy-land/chain"] = Pair.prototype.chain;
Pair.prototype["fantasy-land/of"] = Pair.prototype.of;
Pair.prototype["fantasy-land/ap"] = Pair.prototype.ap;

var Pair$1 = function Pair$1(v1, v2) {
  return new Pair(v1, v2);
};
Pair$1.of = Pair.prototype.of;
Pair$1.prototype["fantasy-land/of"] = Pair.prototype.of;

function id(x) {
    return x;
}

var Reader = function () {
    function Reader(fn) {
        if (!isFunction(fn)) throw Error("Reader: Must wrap a function");
        this._value = fn;
    }
    Reader.of = function (v) {
        return new Reader(function () {
            return v;
        });
    };
    Reader.prototype.of = function (v) {
        return new Reader(function () {
            return v;
        });
    };
    Reader.prototype.ask = function (fn) {
        if (!fn) return new Reader(id);
        if (isFunction(fn)) return new Reader(fn);
        throw Error("Reader: No argument or function required");
    };
    Reader.prototype.runWith = function (ctx) {
        return this._value(ctx);
    };
    Reader.prototype.ap = function (j) {
        if (!(j instanceof Reader)) throw Error("Reader: Reader required");
        var runWith = this.getValue();
        return new Reader(function (e) {
            var fn = runWith(e);
            if (!isFunction(fn)) throw Error("Reader: Wrapped function must return a function");
            return j.map(fn).runWith(e);
        });
    };
    Reader.prototype.getValue = function () {
        return this._value;
    };
    Reader.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("Reader: Expected a function");
        var runWith = this.getValue();
        return new Reader(function (x) {
            return f(runWith(x));
        });
    };
    Reader.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("Reader: Expected a function");
        var runWith = this.getValue();
        return new Reader(function (e) {
            var m = f(runWith(e));
            if (!(m instanceof Reader)) throw Error("Reader: Function must return a Reader");
            return m.runWith(e);
        });
    };
    Reader.prototype.toString = function () {
        var v = this.getValue();
        return "Reader(" + v + ")";
    };
    return Reader;
}();
Reader.prototype["fantasy-land/map"] = Reader.prototype.map;
Reader.prototype["fantasy-land/chain"] = Reader.prototype.chain;
Reader.prototype["fantasy-land/of"] = Reader.prototype.of;
Reader.prototype["fantasy-land/ap"] = Reader.prototype.ap;

var Reader$1 = function Reader$1(v) {
  return new Reader(v);
};
Reader$1.of = Reader.prototype.of;
Reader$1.ask = Reader.prototype.ask;
Reader$1.prototype["fantasy-land/of"] = Reader.prototype.of;

var State = function () {
    function State(fn) {
        if (!isFunction(fn)) throw Error("State: Must wrap a function");
        this._value = fn;
    }
    State.of = function (v) {
        return new State(function (s) {
            return Pair$1(v, s);
        });
    };
    State.prototype.of = function (v) {
        return new State(function (s) {
            return Pair$1(v, s);
        });
    };
    State.prototype.get = function (fn) {
        if (!fn) return new State(function (s) {
            return Pair$1(s, s);
        });
        if (isFunction(fn)) return new State(function (s) {
            return Pair$1(fn(s), s);
        });
        throw Error("State: No argument or function required");
    };
    State.prototype.put = function (s) {
        if (s) return new State(function (_) {
            return Pair$1(s, _);
        });
        throw Error("State: No argument or function required");
    };
    State.prototype.runWith = function (s) {
        var p = this._value(s);
        if (!p.fst || !p.snd) throw Error("State: Must wrap a function of the form s -> Pair a s");
        return p;
    };
    State.prototype.execWith = function (s) {
        var pair = this.runWith(s);
        return pair.snd();
    };
    State.prototype.evalWith = function (s) {
        var pair = this.runWith(s);
        return pair.fst();
    };
    State.prototype.ap = function (j) {
        if (!(j instanceof State)) throw Error("State: State required");
        var runWith = this._value;
        return new State(function (e) {
            var p = runWith(e);
            if (!p.fst || !p.snd) throw Error("State: Must wrap a function of the form s -> Pair a s");
            var fn = p.fst();
            if (!isFunction(fn)) throw Error("State: Source value must be a function");
            return j.map(fn).runWith(p.snd());
        });
    };
    State.prototype.getValue = function () {
        return this._value;
    };
    State.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("State: Expected a function");
        var runWith = this._value;
        return new State(function (s) {
            var m = runWith(s);
            return Pair$1(f(m.fst()), m.snd());
        });
    };
    State.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("State: Expected a function");
        var runWith = this.getValue();
        return new State(function (s) {
            var p = runWith(s);
            var m = f(p.fst());
            if (!(m instanceof State)) throw Error("State: Function must return a State");
            return m.runWith(p.snd());
        });
    };
    State.prototype.toString = function () {
        var v = this.getValue();
        return "State(" + v + ")";
    };
    return State;
}();
State.prototype["fantasy-land/map"] = State.prototype.map;
State.prototype["fantasy-land/chain"] = State.prototype.chain;
State.prototype["fantasy-land/of"] = State.prototype.of;
State.prototype["fantasy-land/ap"] = State.prototype.ap;

var State$1 = function State$1(v) {
  return new State(v);
};
State$1.of = State.prototype.of;
State$1.get = State.prototype.get;
State$1.put = State.prototype.put;
State$1.prototype["fantasy-land/of"] = State.prototype.of;

function compose() {
    var fns = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        fns[_i] = arguments[_i];
    }
    if (fns.length <= 0) throw Error("Expected a function to compose");
    return fns.reduce(function (f, g) {
        return function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return f(g.apply(void 0, args));
        };
    });
}

function once(f) {
    if (!isFunction(f)) throw Error("Function not provided");
    var _called = false;
    var _result;
    return function () {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        if (!_called) {
            _called = true;
            _result = f.apply(null, args);
        }
        return _result;
    };
}

var Task = function () {
    function Task(f) {
        if (isFunction(f)) this._value = f;else throw Error("Task: Expected a Function");
    }
    Task.of = function (v) {
        return new Task(function (_, resolve) {
            return resolve(v);
        });
    };
    Task.rejected = function (v) {
        return new Task(function (reject, _) {
            return reject(v);
        });
    };
    Task.prototype.of = function (v) {
        return new Task(function (_, resolve) {
            return resolve(v);
        });
    };
    Task.prototype.rejected = function (v) {
        return new Task(function (reject, _) {
            return reject(v);
        });
    };
    Task.prototype.fork = function (reject, resolve) {
        if (!isFunction(resolve) || !isFunction(reject)) throw Error("Task: Reject and Resolve need to be functions");
        var fn = this.getValue();
        fn(reject, resolve);
    };
    Task.prototype.toString = function () {
        var fork = this.getValue();
        return "Task(" + fork.name + ")";
    };
    Task.prototype.map = function (f) {
        if (!isFunction(f)) throw Error("Task: Expected a function");
        var fork = this.getValue();
        return new Task(function (rej, res) {
            return fork(rej, compose(res, f));
        });
    };
    Task.prototype.getValue = function () {
        return this._value;
    };
    Task.prototype.ap = function (t) {
        if (!(t instanceof Task)) throw Error("Task: type mismatch");
        var thisFork = this.getValue();
        var value;
        var fn;
        var gotValues = false;
        var gotFuction = false;
        var rejected = false;
        return new Task(function (rej, res) {
            var rejOnce = compose(function () {
                rejected = true;
            }, once(rej));
            var resolveBoth = function resolveBoth() {
                if (gotValues && gotFuction && !rejected) {
                    var exec = compose(res, fn);
                    exec.apply(null, value);
                }
            };
            thisFork(rejOnce, function () {
                var values = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    values[_i] = arguments[_i];
                }
                value = values;
                gotValues = true;
                resolveBoth();
            });
            t.fork(rejOnce, function (f) {
                if (!isFunction(f)) throw Error("Task: Wrapped value should be a function");
                fn = f;
                gotFuction = true;
                resolveBoth();
            });
        });
    };
    Task.prototype.concat = function (t) {
        if (!(t instanceof Task)) throw Error("Task: type mismatch");
        var thisFork = this.getValue();
        var thatFork = t.getValue();
        return new Task(function (rej, res) {
            var rejected = false;
            var rejOnce = compose(function () {
                rejected = true;
            }, once(rej));
            var result1;
            var result2;
            var resolveBoth = function resolveBoth() {
                if (result1 && result2 && !rejected) res.apply(null, __spreadArrays(result1, result2));
            };
            thisFork(rejOnce, function () {
                var values = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    values[_i] = arguments[_i];
                }
                result1 = values;
                resolveBoth();
            });
            thatFork(rejOnce, function () {
                var values = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    values[_i] = arguments[_i];
                }
                result2 = values;
                resolveBoth();
            });
        });
    };
    Task.prototype.chain = function (f) {
        if (!isFunction(f)) throw Error("Task: Function required");
        var thisFork = this.getValue();
        return new Task(function (rej, res) {
            thisFork(rej, function () {
                var args = [];
                for (var _i = 0; _i < arguments.length; _i++) {
                    args[_i] = arguments[_i];
                }
                var t = f.call(null, args);
                if (!t.fork) throw Error("Task: function must return another Task");
                t.fork(rej, res);
            });
        });
    };
    Task.prototype.toPromise = function () {
        var thisFork = this.getValue();
        return new Promise(function (res, rej) {
            thisFork(rej, res);
        });
    };
    return Task;
}();
Task.prototype["fantasy-land/map"] = Task.prototype.map;
Task.prototype["fantasy-land/concat"] = Task.prototype.concat;
Task.prototype["fantasy-land/chain"] = Task.prototype.chain;
Task.prototype["fantasy-land/of"] = Task.prototype.of;
Task.prototype["fantasy-land/ap"] = Task.prototype.ap;

var Task$1 = function Task$1(f) {
  return new Task(f);
};
Task$1.of = Task.of;
Task$1.rejected = Task.rejected;

function _bimap(f1, f2, b) {
    if (!isFunction(f1) || !isFunction(f2)) throw Error("Expected functions");
    if (!b.bimap) throw Error("Expected a BiFunctor as last argument");
    return b.bimap(f1, f2);
}
var bimap = curry(_bimap);

function chain(f, m) {
    if (!m.chain) throw Error("chain not implemented");
    if (!isFunction(f)) throw Error("function not provided");
    return m.chain.call(m, f);
}
var chain$1 = curry(chain);

function constant(x) {
    return function constant_() {
        return x;
    };
}

function fmap(fn, f) {
    if (!isFunction(fn)) throw Error("function not provided");
    if (!f.map) throw Error("Functor not found");
    return f.map.call(f, fn);
}
var fmap$1 = curry(fmap);

function liftA2(f, f1, f2) {
    var res = f1.map(f).ap(f2);
    return res;
}
var liftA2$1 = curry(liftA2);

function liftA3(f, f1, f2, f3) {
    var res = f1.map(f).ap(f2).ap(f3);
    return res;
}
var liftA3$1 = curry(liftA3);

function liftAn(f, fns) {
    if (!isFunction(f)) throw Error("Expected a function");
    var init = fns[0].map(f);
    var res = init;
    if (fns.length > 1) {
        var rest = fns.slice(1);
        res = rest.reduce(function applyFn(a, ca) {
            return a.ap(ca);
        }, init);
    }
    return res;
}
var liftAn$1 = curry(liftAn);

function ocurry(fn, args) {
    if (!isFunction(fn)) throw Error("Expected a function as first argument to ocurry");
    if (fn.length > 1) throw Error("Function passed as first argument should not take more than one argument. Argument should be a dictionary");
    return function ocurried(ar) {
        var curArgs = Object.keys(ar);
        var diff = args.filter(function (x) {
            return curArgs.indexOf(x) < 0;
        });
        if (diff.length > 0) return function (ar2) {
            return ocurried.call(null, Object.assign({}, ar, ar2));
        };
        return fn.call(null, ar);
    };
}

var index = {
    Either: Either,
    Maybe: Maybe,
    Pair: Pair$1,
    Reader: Reader$1,
    State: State$1,
    Task: Task$1,
    bimap: bimap,
    caseOf: caseOf$1,
    chain: chain$1,
    compose: compose,
    constant: constant,
    curry: curry,
    eitherToMaybe: eitherToMaybe,
    fmap: fmap$1,
    id: id,
    liftA2: liftA2$1,
    liftA3: liftA3$1,
    liftAn: liftAn$1,
    maybeToEither: maybeToEither,
    ocurry: ocurry,
    once: once,
    prop: prop
};

module.exports = index;
